import gc
import itertools
import logging
import sys
import time
import numpy as np
#from mapper import read_fasta_into_chunks, \

from kmer_mapper.mapper import map_kmers_to_graph_index
from kmer_mapper.util import remap_array
from graph_kmer_index import KmerIndex
from shared_memory_wrapper import from_shared_memory, to_shared_memory, SingleSharedArray, object_to_shared_memory
from shared_memory_wrapper.shared_memory import get_shared_pool, close_shared_pool, object_from_shared_memory
from itertools import repeat
from graph_kmer_index.collision_free_kmer_index import MinimalKmerIndex

from .hash_table import NodeCount
from .parser import OneLineFastaParser, Sequences, OneLineFastaParser2bit, BufferedNumpyParser, OneLineFastaBuffer2Bit
from .util import log_memory_usage_now
from .kmers import KmerHash, TwoBitHash
from npstructures import Counter
from bionumpy.kmers import fast_hash
import bionumpy as bnp


def map_kmers_to_graph_index_wrapper(*args):
    return map_kmers_to_graph_index(*args)


def convert_byte_read_array_to_int_array(reads, read_length):
    # each item is a byesequence, one byte for each base pair
    return np.frombuffer(reads, dtype=np.uint8).reshape((len(reads), read_length))


def convert_read_matrix_to_numeric(read_matrix, give_complement_base_values=False):
    # from byte values
    from_values = np.array([65, 67, 71, 84, 97, 99, 103, 116], dtype=np.uint64)  # NB: Must be increasing
    # to internal base values for a, c, t, g
    to_values = np.array([0, 1, 3, 2, 0, 1, 3, 2], dtype=np.uint64)
    if give_complement_base_values:
        to_values = np.array([2, 3, 1, 0, 2, 3, 1, 0], dtype=np.uint64)

    return remap_array(read_matrix, from_values, to_values)


def get_kmer_hashes(numeric_read_matrix, is_complement=False, k=31):
    power_array = np.power(4, np.arange(0, k, dtype=np.uint64), dtype=np.uint64)
    if is_complement:
        power_array = np.power(4, np.arange(0, k, dtype=np.uint64)[::-1], dtype=np.uint64)

    k_half = k//2  # we don't want first k half and last k half columns
    assert numeric_read_matrix.dtype == np.uint64
    assert power_array.dtype == np.uint64

    #res = scipy.signal.fftconvolve(numeric_read_matrix, power_array, mode="valid")[:,k_half:-k_half]
    return convolve1d(numeric_read_matrix, power_array, mode="constant")[:,k_half:-k_half]

def get_kmer_hashes_numpy(numeric_read_matrix, is_complement=False, k=31):
    power_array = np.power(4, np.arange(0, k, dtype=np.uint64), dtype=np.uint64)
    if is_complement:
        power_array = np.power(4, np.arange(0, k, dtype=np.uint64)[::-1], dtype=np.uint64)

    assert numeric_read_matrix.dtype == np.uint64
    assert power_array.dtype == np.uint64

    # idea: convolve a 1d version of the 2d matrix. Use mode full and remove the first k-1 which are when power array is
    # convolved outside beginning. Reshape back to matrix and remove last k-1 columns which are overlaps between rows
    function = np.convolve
    #function = scipy.signal.convolve
    flat_numeric_read_matrix = numeric_read_matrix.flatten()
    if len(flat_numeric_read_matrix) == 0:
        logging.warning("Read matrix is empty")
        raise Exception("No reads")
        return np.zeros_like(numeric_read_matrix)

    return function(flat_numeric_read_matrix, power_array, mode="full")[k-1:].reshape(numeric_read_matrix.shape)[:,:-(k-1)]


def get_unique_kmers_and_counts(hash_matrix):
    return np.unique(hash_matrix, return_counts=True)
    unique = pd.value_counts(hash_matrix, sort=False)
    return unique.keys(), unique.values


def get_kmers_from_read_matrix(read_matrix, mask, k=31, return_only_kmers=False, include_reverse_complement=True):
    logging.info("k=%d" % k)
    t = time.time()
    numeric_reads = convert_read_matrix_to_numeric(read_matrix)
    numeric_reads_complement = (numeric_reads + 2) % 4  #convert_read_matrix_to_numeric(read_matrix, True)
    logging.info("Time to convert reads to numeric: %.3f" % (time.time()-t))
    t = time.time()

    mask = mask[:,k-1:]  # make mask for hashes (we don't want hashes for bases not from reads, which are False in original mask)
    hashes = get_kmer_hashes_numpy(numeric_reads, k=k)[mask]
    if include_reverse_complement:
        hashes_complement = get_kmer_hashes_numpy(numeric_reads_complement, is_complement=True, k=k)[mask]
    logging.info("Time to get kmer hashes: %.3f" % (time.time()-t))
    t = time.time()

    if include_reverse_complement:
        all_hashes = np.concatenate([hashes, hashes_complement])
    else:
        all_hashes = hashes  # np.concatenate([hashes, hashes_complement])

    logging.info("N hashes in total: %d" % len(all_hashes))
    logging.info("Time to concatenate hashes and complement hashes: %.3f" % (time.time()-t))
    t = time.time()

    if return_only_kmers:
        return all_hashes

    unique, counts = get_unique_kmers_and_counts(all_hashes)
    #logging.info("N unique hashes  : %d" % len(unique))
    logging.info("Time to count kmers: %.3f" % (time.time()-t))

    return unique, counts


def map_fasta_single_thread_with_numpy_parsing(data):
    logging.info("---------------------")
    logging.info("Starting processing chunk of reads")
    log_memory_usage_now()
    time_start = time.perf_counter()
    reads, args = data

    logging.info("Reeading chunk of reads from shared memory")
    raw_chunk = from_shared_memory(args.buffer_type, reads)
    sequence_chunk = raw_chunk.get_sequences()

    if not args.use_numpy:
        node_counts = from_shared_memory(SingleSharedArray, "counts_shared"+args.random_id).array

    t = time.perf_counter()
    hashes = TwoBitHash(k=args.kmer_size).get_kmer_hashes(sequence_chunk)
    logging.info("Time spent to get %d kmer hashes: %.3f" % (len(hashes), time.perf_counter() - t))

    t = time.perf_counter()
    if args.use_numpy:
        counter = object_from_shared_memory(args.counter)
        t_before_count = time.perf_counter()
        counter.counter.count(hashes.astype(np.int64))
        logging.info("Done counting, took %.3f sec" % (time.perf_counter()-t_before_count))
    else:
        logging.info("Using cython counter")
        kmer_index = from_shared_memory(KmerIndex, args.kmer_index)
        node_counts += map_kmers_to_graph_index(kmer_index, args.max_node_id, hashes, args.max_hits_per_kmer)

    logging.info("Getting node counts took %.3f sec" % (time.perf_counter()-t))
    logging.info("Done with chunk. Took %.3f sec" % (time.perf_counter()-time_start))


def map_fasta(args, kmer_index):
    logging.info("Mapping fasta to kmer index %s" % args.kmer_index)
    args.random_id = str(np.random.random())
    logging.info("Putting kmer index in shared memory")
    if not args.use_numpy:
        args.kmer_index = to_shared_memory(kmer_index)
    else:
        args.kmer_index = kmer_index

    logging.info("Done putting in shared memory")

    start_time = time.time()

    pool = get_shared_pool(args.n_threads)

    if not args.use_numpy:
        node_counts = np.zeros(args.max_node_id+1, dtype=np.uint32)
        to_shared_memory(SingleSharedArray(node_counts), "counts_shared" + args.random_id)
        node_counts = None


    parser = BufferedNumpyParser.from_filename(args.fasta_file, args.chunk_size)
    args.buffer_type = OneLineFastaBuffer2Bit
    chunks = parser.get_chunks()
    reads = (to_shared_memory(chunk) for chunk in chunks)
    func = map_fasta_single_thread_with_numpy_parsing

    i = 0
    data = zip(reads, repeat(args))

    logging.info(func)
    logging.info(args)

    for result in pool.imap(func, data, chunksize=1):
    #for result in map(func, data):
        logging.info("Done with %d chunks" % i)
        i += 1

    if args.use_numpy:
        t0 = time.perf_counter()
        counter = object_from_shared_memory(args.counter)
        kmer_counts = counter.counter[counter.kmers]
        logging.info("Time to get counts for the index kmers from counter: %.3f" % (time.perf_counter()-t0))
        node_counts = np.bincount(counter.nodes, kmer_counts)
        logging.info("Sum of node counts: %d" % np.sum(node_counts))
        logging.info("Sum of counts: %d" % np.sum(kmer_counts))
    else:
        node_counts = from_shared_memory(SingleSharedArray, "counts_shared"+args.random_id).array            
    
    np.save(args.output_file, node_counts)
    logging.info("Saved node counts to %s.npy" % args.output_file)
    logging.info("Spent %.3f sec in total mapping kmers using %d threads" % (time.time()-start_time, args.n_threads))

    close_shared_pool()
    
from bionumpy.kmers import fast_hash
from bionumpy.kmers import KmerEncoding


def _parallel_sequence_map_wrapper(data):
    index_id, max_node_id, k, sequence_id = data
    sequence = object_from_shared_memory(sequence_id)
    sequence = sequence.ravel()  # .astype(np.int64)
    #hasher = KmerEncoding(k, None, 4)
    #hashes = hasher.rolling_window(sequence)
    #kmers = hashes.astype(np.uint64)
    t0 = time.perf_counter()
    kmers = fast_hash(sequence[::-1], k).astype(np.uint64)
    logging.info("Hashing %d kmers took %.4f sec" % (len(kmers), time.perf_counter()-t0))
    index = object_from_shared_memory(index_id)


    t0 = time.perf_counter()
    mapped = map_kmers_to_graph_index(index, max_node_id, kmers)
    logging.info("Mapping %d kmers took %.4f sec" % (len(kmers), time.perf_counter()-t0))
    return mapped


def _parallel_map_wrapper(data):
    index_id, max_node_id, kmer_id = data
    kmers = object_from_shared_memory(kmer_id)
    index = object_from_shared_memory(index_id)
    return map_kmers_to_graph_index(index, max_node_id, kmers)


class ParalellMapper:
    def __init__(self, kmer_index, max_node_id, n_threads=8):
        t = time.perf_counter()
        self._max_node_id = max_node_id
        self._kmer_mapper = kmer_index
        self._kmer_mapper = object_to_shared_memory(kmer_index)
        self._n_threads = n_threads
        self._counts = np.zeros(max_node_id, dtype=np.uint32)
        self._pool = get_shared_pool(self._n_threads)
        logging.info("Time spent initing ParallellMapper: %.4f" % (time.perf_counter()-t))

    def map_reads_file(self, read_file_name, k, chunk_size=10000000):
        chunks = bnp.open(read_file_name, chunk_size=chunk_size)
        #chunks = (chunk.sequence.ravel() for chunk in chunks)
        sequences = (object_to_shared_memory(chunk.sequence) for chunk in chunks)

        for result in self._pool.imap(_parallel_sequence_map_wrapper,
                                  zip(itertools.repeat(self._kmer_mapper), itertools.repeat(self._max_node_id - 1),
                                      itertools.repeat(k),
                                      sequences)):
            self._counts += result

    def map_numeric_sequence(self, numeric_sequence, k):
        t = time.perf_counter()
        assert numeric_sequence.dtyp == np.uint8
        sequences = np.array_split(numeric_sequence, self._n_threads)
        logging.info("time spent splitting sequences: %.3f" % (time.perf_counter()-t))

        t = time.perf_counter()
        sequences = [object_to_shared_memory(k) for k in sequences]
        func = self._pool.imap
        if self._n_threads == 1:
            func = map

        logging.info("time spent writing sequences to shared memory: %.3f" % (time.perf_counter()-t))
        for result in func(_parallel_sequence_map_wrapper,
                                      zip(itertools.repeat(self._kmer_mapper), itertools.repeat(self._max_node_id - 1), itertools.repeat(k),
                                          sequences)):
            self._counts += result

    def map(self, kmers):
        assert kmers.dtype == np.uint64
        kmers = np.array_split(kmers, self._n_threads)
        kmers = (object_to_shared_memory(k) for k in kmers)
        for result in self._pool.imap(_parallel_map_wrapper,
                                      zip(itertools.repeat(self._kmer_mapper), itertools.repeat(self._max_node_id-1), kmers)):
            self._counts += result

    def get_results(self):
        #logging.info("Sum of counts: %d" % np.sum(self._counts))
        return self._counts

    def reset(self):
        self._counts[:] = 0


from .encodings import ACTGTwoBitEncoding, BaseEncoding

class Sequences:
    def __init__(self, sequences, intervals, encoding=BaseEncoding):
        assert intervals[0].size == intervals[1].size
        self.sequences = sequences
        self.intervals = intervals
        self.intervals_start = intervals[0]
        self.intervals_end = intervals[1]
        self.encoding = encoding

    def __len__(self):
        return len(self.offsets)+1

    def __getitem__(self, i):
        start, end = (0, len(self.sequences))
        if i > 0:
            start = self.offsets[i-1]
        if i < len(self.offsets):
            end = self.offsets[i]
        return self.sequences[start:end]

    def __repr__(self):
        return "Seqs(%s, %s)" % (str(self.sequences), self.intervals)

    def nbytes(self):
        return self.sequences.nbytes + self.intervals_start.nbytes + self.intervals_end.nbytes





import logging
from .encodings import twobit_swap, ACTGTwoBitEncoding
import numpy as np

class KmerLookup:
    def __init__(self, kmers, representative_kmers, lookup):
        self._kmers = kmers
        self._representative_kmers = representative_kmers
        self._lookup = lookup

    def max_node_id(self):
        return np.max(self._lookup)

    def to_file(self, file_name):
        np.savez(file_name, 
                 kmers = self._kmers,
                 representative_kmers = self._representative_kmers,
                 lookup = self._lookup)

    @classmethod
    def from_file(cls, file_name):
        data = np.load(file_name)
        return cls(data["kmers"],
                   data["representative_kmers"],
                   data["lookup"])

    def index_kmers(self):
        self._kmers.sort()
        self._representative_kmers = np.searchsorted(self._kmers, self._representative_kmers)

    def _get_indexes(self, kmers):
        return np.searchsorted(self._kmers, kmers)

    def count_kmers(self, kmers):
        indexes = self._get_indexes(kmers)
        return np.bincount(indexes[kmers==self._kmers[np.minimum(indexes, self._kmers.size-1)]],
                           minlength=self._kmers.size)
        
    def get_node_counts(self, kmers):
        counts = self.count_kmers(kmers)
        cum_counts = np.insert(np.cumsum(counts[self._representative_kmers]), 0, 0)
        node_counts = cum_counts[self._lookup[:, 1]]-cum_counts[self._lookup[:, 0]]
        return node_counts



class SimpleKmerLookup(KmerLookup):
    def get_node_counts(self, kmers):
        counts = self.count_kmers(kmers)
        return np.bincount(self._lookup, counts[self._representative_kmers])

    @classmethod
    def from_old_index_files(cls, filename):
        logging.info("From old index files")
        data = np.load(filename)
        kmers = data["kmers"]
        unique_kmers = np.unique(kmers)
        k = cls(unique_kmers, kmers, data["nodes"])
        k.index_kmers()
        return k


class IndexedSortedLookup(SimpleKmerLookup):
    def __init__(self, sorted_values, index, mod, row_sizes):
        self._sorted_values = np.append(sorted_values, np.array([4**31], dtype=np.uint64))
        self._mod = mod
        self._index = index
        self._row_sizes = row_sizes
        # THis is maybe to memory demanding
        # self._boundries = np.vstack((self._index, self._index+self._row_sizes-1))

    @classmethod
    def from_values_and_mod(cls, values, mod, max_value, do_sort=False):
        if do_sort:
            values = np.sort(values)
        m = values//mod
        index = np.searchsorted(m, np.arange(max_value//mod+2))
        row_sizes = index[1:]-index[:-1]
        return cls(values, index[:-1], mod, row_sizes)

    def _get_matched_data(self, row_numbers):
        row_sizes = self._row_sizes[row_numbers]
        mask = np.flatnonzero(row_sizes)# x > 0
        row_sizes = row_sizes[mask]
        L = self._index[row_numbers[mask]]
        R = L + row_sizes
        return L, R, mask, np.max(row_sizes)


    def find_matches(self, queries):
        L, R, mask, row_length = self._get_matched_data(queries // self._mod)
        n_iter = int(np.log2(row_length))+1
        queries = queries[mask]
        found_indices = self._binary_search(queries, L, R, n_iter)
        return found_indices[self._sorted_values[found_indices]==queries]

    def _binary_search(self, queries, L, R, n_iter):
        """
        https://en.wikipedia.org/wiki/Binary_search_algorithm#Alternative_procedure
        """
        boundries = np.vstack((L, R))
        indexes = np.arange(boundries.shape[-1])
        for _ in range(n_iter):
            m = (boundries.sum(axis=0))//2
            is_larger = (self._sorted_values[m] > queries).view(np.uint8)
            boundries[is_larger, indexes] = m
            """
            m = (L+R+1)//2
            is_larger = self._sorted_values[m] > queries
            R[is_larger] = m[is_larger]-1
            L[~is_larger] = m[~is_larger]
            """
        return boundries[0] 
                          
    def find_queries(self, queries):
        row_numbers = queries // self._mod
        L = self._index[row_numbers]
        row_sizes = self._row_sizes[row_numbers]
        R = L + row_sizes-1
        m = L
        n_iter = int(np.log2(np.max(row_sizes)))+1
        for _ in range(n_iter):
            m = (L+R+1)//2
            is_larger = self._sorted_values[m] > queries
            R[is_larger] = m[is_larger]-1
            L[~is_larger] = m[~is_larger]
        return L

        
class ShortCircuitIndexedSortedLookup(IndexedSortedLookup):
    def _get_matched_data(self, row_numbers):
        row_sizes = self._row_sizes[row_numbers]
        mask = np.flatnonzero(row_sizes)#  > 0
        row_sizes = row_sizes[mask]
        n_iters = np.floor(np.log2(row_sizes)).astype(int)+1
        args = np.argsort(n_iters)
        mask = mask[args]
        L = self._index[row_numbers[mask]]
        R = L + row_sizes[args]
        return L, R, mask, n_iters


    def find_matches(self, queries):
        L, R, mask, n_iters = self._get_matched_data(queries // self._mod)
        queries = queries[mask]
        found_indices = self._binary_search(queries, L, R, n_iters)
        return found_indices[self._sorted_values[found_indices]==queries]

    def _binary_search(self, queries, L, R, n_iters):
        """
        https://en.wikipedia.org/wiki/Binary_search_algorithm#Alternative_procedure
        """
        boundries = np.vstack((L, R))
        indexes = np.arange(boundries.shape[-1])
        # args = np.argsort(n_iters)
        #boundries = _boundries[:, args]
        #queries = _queries[args]
        stops = np.cumsum(np.bincount(n_iters))
        for stop in stops[:-1]:
            m = (boundries[:, stop:].sum(axis=0))//2
            is_larger = (self._sorted_values[m] > queries[stop:]).view(np.uint8)
            boundries[is_larger, indexes[stop:]] = m
            """
            m = (L+R+1)//2
            is_larger = self._sorted_values[m] > queries
            R[is_larger] = m[is_larger]-1
            L[~is_larger] = m[~is_larger]
            """
        return boundries[0]


class AdvancedKmerLookup(SimpleKmerLookup):
    n_bins = 200000000
    max_value = 4**31

    def _get_indexes(self, kmers):
        return self._indexed_lookup.find_queries(kmers)

    def index_kmers(self):
        self._kmers.sort()
        self._representative_kmers = np.searchsorted(self._kmers, self._representative_kmers)
        lookp_class = IndexedSortedLookup
        self._indexed_lookup = lookup_class.from_values_and_mod(
            self._kmers, self.max_value//self.n_bins+1,
            self.max_value, do_sort=False)

class Advanced2(AdvancedKmerLookup):
    lookup_class = IndexedSortedLookup
    def count_kmers(self, kmers):
        indices = self._indexed_lookup.find_matches(kmers)
        return np.bincount(indices, minlength=self._kmers.size)

    @classmethod
    def from_file(cls, filename):
        D = np.load(filename)
        obj = cls(D["kmers"], D["representative_kmers"], D["lookup"])
        obj._indexed_lookup=cls.lookup_class(
            D["kmers"], D["index"], D["mod"], D["row_size"])
        return obj

    def to_file(self, filename):
        np.savez(filename, 
                 kmers=self._kmers,
                 representative_kmers=self._representative_kmers,
                 lookup=self._lookup,
                 index=self._indexed_lookup._index,
                 mod=self._indexed_lookup._mod,
                 row_size=self._indexed_lookup._row_sizes)
    

class NewHashLookup(Advanced2):
    dtype=np.uint64
    k=31
    lookup_class = ShortCircuitIndexedSortedLookup
    # lookup_class = IndexedSortedLookup
    @classmethod
    def _rehash_kmers(cls, kmers):
        new_hashes = twobit_swap(kmers)>>cls.dtype(2)
        reverse_hashes = ACTGTwoBitEncoding.complement(kmers) & cls.dtype(4**cls.k-1)
        return np.concatenate((new_hashes, reverse_hashes))

    @classmethod
    def from_lookup_with_old_hash(cls, old_lookup):
        kmers = cls._rehash_kmers(old_lookup._representative_kmers)
        nodes = np.concatenate((old_lookup._lookup, old_lookup._lookup))
        ret = cls(np.unique(kmers), kmers, nodes)
        ret.index_kmers()
        return ret

    @classmethod
    def from_old_index_files(cls, filename):
        logging.info("From old index files")
        data = np.load(filename)
        kmers = cls._rehash_kmers(data["kmers"])
        unique_kmers = np.unique(kmers)
        nodes = data["nodes"]
        nodes = np.concatenate((nodes, nodes))
        k = cls(unique_kmers, kmers, nodes)
        k.index_kmers()
        return k

class ModuloHashLookup(ShortCircuitIndexedSortedLookup):

    def find_matches(self, queries):
        hashes = queries % self._mod
        row_sizes = self._row_sizes[hashes]
        mask = np.flatnonzero(row_sizes)
        row_sizes = row_sizes[mask]
        n_iters = np.floor(np.log2(row_sizes)).astype(int)+1
        args = np.argsort(n_iters)
        mask = mask[args]
        R = self._index[hashes[mask]]
        L = R - row_sizes[args]
        queries = queries[mask]
        found_indices = self._binary_search(queries, L, R, n_iters)
        return found_indices[self._sorted_values[found_indices]==queries]

    def get_index(self, kmers):
        row_numbers = kmers % self._mod
        row_sizes = self._row_sizes[row_numbers]
        assert np.min(row_sizes)>0, np.min(row_sizes)
        R = self._index[row_numbers]
        L = R - row_sizes
        n_iters = np.floor(np.log2(row_sizes)).astype(int)+1
        n_iters = np.maximum(n_iters, np.max(n_iters))# np.floor(np.log2(row_sizes)).astype(int)+1
        found_indices = self._binary_search(kmers, L, R, n_iters)
        return found_indices

    @classmethod
    def from_values_and_mod(cls, kmers, modulo):
        hashes = (kmers % modulo).astype(int)
        counts = np.bincount(hashes, minlength=modulo)
        args = np.lexsort([kmers, hashes])
        hash_sorted = kmers[args]
        row_sizes=counts
        index = np.cumsum(counts)
        return cls(hash_sorted, index, modulo, row_sizes)
        
class NewHashLookup2(NewHashLookup):
    lookup_class = ModuloHashLookup
    def index_kmers(self):
        self._indexed_lookup = self.lookup_class.from_values_and_mod(
            self._kmers, np.uint64(self.n_bins))
        self._representative_kmers = self._indexed_lookup.get_index(self._representative_kmers)

    @classmethod
    def from_old_file(cls, filename):
        D = np.load(filename)
        kmers = D["kmers"]
        repr_kmers = kmers[D["representative_kmers"]]
        obj = cls(kmers, repr_kmers, D["lookup"])
        obj.index_kmers()
        return obj

    @classmethod
    def from_file(cls, filename):
        D = np.load(filename)
        obj = cls(D["kmers"], D["representative_kmers"], D["lookup"])
        obj._indexed_lookup=cls.lookup_class(
            D["kmers"], D["index"], D["mod"], D["row_size"])
        return obj

    def to_file(self, filename):
        np.savez(filename, 
                 kmers=self._indexed_lookup._sorted_values,
                 representative_kmers=self._representative_kmers,
                 lookup=self._lookup,
                 index=self._indexed_lookup._index,
                 mod=self._indexed_lookup._mod,
                 row_size=self._indexed_lookup._row_sizes)



import numpy as np
import logging

class ModuloHashLookup:
    def __init__(self, values, mod, n_entries, lookup_end):
        self._values = values # np.append(sorted_values, np.array([4**31], dtype=np.uint64))
        self._mod = mod
        self._n_entries = n_entries
        self._lookup_end = lookup_end

    def _get_hash(self, queries):
        return queries % self._mod

    def get_hit_idxs(self, queries):
        hashes = self._get_hash(queries)
        return np.flatnonzero(self._n_entries[hashes])

    def get_hits(self, queries):
        queries = queries[self.get_hit_idxs(queries)]
        indexes = self.get_index(queries)
        return indexes[self._values[indexes]==queries]

    def get_index(self, queries):
        hashes = self._get_hash(queries)
        n_entries = self._n_entries[hashes]
        n_iters = np.floor(np.log2(n_entries)).astype(int)+1
        n_iter_sort_args = np.argsort(n_iters)
        n_entries = n_entries[n_iter_sort_args]
        hashes = hashes[n_iter_sort_args]
        n_iters = n_iters[n_iter_sort_args]
        R = self._lookup_end[hashes]
        L = R - n_entries
        found_indices = np.zeros_like(queries, dtype=int)
        found_indices[n_iter_sort_args] = self._binary_search(queries[n_iter_sort_args], L, R, n_iters)
        return found_indices

    def _binary_search(self, queries, L, R, n_iters):
        """
        https://en.wikipedia.org/wiki/Binary_search_algorithm#Alternative_procedure
        Look for queries in range [L, R>
        Runs only reuqired iterations for each group
        """
        boundries = np.vstack((L, R))
        indexes = np.arange(boundries.shape[-1])
        stops = np.cumsum(np.bincount(n_iters))
        for stop in stops[:-1]:
            m = (boundries[:, stop:].sum(axis=0))//2
            is_larger = (self._values[m] > queries[stop:]).view(np.uint8)
            boundries[is_larger, indexes[stop:]] = m
        return boundries[0]

    @classmethod
    def from_values_and_mod(cls, keys, modulo):
        hashes = (keys % modulo).astype(int)
        counts = np.bincount(hashes, minlength=modulo)
        args = np.lexsort([keys, hashes])
        hash_sorted = keys[args]
        n_entries=counts
        lookup_end = np.cumsum(counts)
        return cls(hash_sorted, modulo, n_entries, lookup_end)

class NodeCount:
    dtype=np.uint64
    k=31
    lookup_class = ModuloHashLookup
    n_bins = 200000003

    def __init__(self, kmers, kmer_indexes, node_ids):
        self._kmers = kmers
        self._node_ids = node_ids
        self._kmer_indexes = kmer_indexes

    def max_node_id(self):
        return np.max(self._node_ids)

    def index_kmers(self):
        self._indexed_lookup = self.lookup_class.from_values_and_mod(
            self._kmers, np.uint64(self.n_bins))
        self._kmer_indexes = self._indexed_lookup.get_index(self._kmer_indexes)

    def get_node_counts(self, kmers):
        counts = self.count_kmers(kmers)
        return self.get_node_counts_from_kmer_counts(counts)
        # return np.bincount(self._node_ids, counts[self._kmer_indexes])

    def get_node_counts_from_kmer_counts(self, kmer_counts):
        return np.bincount(self._node_ids, kmer_counts[self._kmer_indexes])

    def count_kmers(self, kmers):
        indices = self._indexed_lookup.get_hits(kmers)
        # indices = self._indexed_lookup.find_matches(kmers)
        return np.bincount(indices, minlength=self._kmers.size)

    @classmethod
    def from_old_file(cls, filename):
        D = np.load(filename)
        kmers = D["kmers"]
        repr_kmers = kmers[D["representative_kmers"]]
        obj = cls(kmers, repr_kmers, D["lookup"])
        obj.index_kmers()
        return obj

    @classmethod
    def from_file(cls, filename):
        D = np.load(filename)
        obj = cls(D["kmers"], D["kmer_indexes"], D["node_ids"])
        obj._indexed_lookup=cls.lookup_class(
            D["kmers"], D["mod"], D["n_entries"], D["lookup_end"])
        return obj

    @classmethod
    def from_old_index_files(cls, filename):
        logging.info("From old index files")
        data = np.load(filename)
        kmers = data["kmers"]
        unique_kmers = np.unique(kmers)
        nodes = data["nodes"]
        k = cls(unique_kmers, kmers, nodes)
        k.index_kmers()
        return k

    def to_file(self, filename):
        np.savez(filename, 
                 kmers=self._indexed_lookup._values,
                 kmer_indexes=self._kmer_indexes,
                 node_ids=self._node_ids,
                 lookup_end=self._indexed_lookup._lookup_end,
                 mod=self._indexed_lookup._mod,
                 n_entries=self._indexed_lookup._n_entries)



import time

import numpy as np
from .parser import get_mask_from_intervals
from .encodings import twobit_swap, ACTGTwoBitEncoding

def get_kmer_mask(intervals, size, k=31):
    starts, ends = intervals
    ends = np.maximum(starts, ends-k+1)
    return get_mask_from_intervals((starts, ends), size-k+1)


class TwoBitHash:
    def __init__(self, k=31, dtype=np.uint64):
        self._dtype = dtype
        self.k = k
        self._mask = dtype(4**k-1)
        self._n_letters_in_dtype = 4*dtype(0).nbytes
        self._shifts = dtype(2)*np.arange(self._n_letters_in_dtype, dtype=dtype)
        self._rev_shifts = self._shifts[::-1]+dtype(2)

    def _get_kmer_hashes(self, sequences):
        """Matching old interface"""
        last_end = sequences.intervals[1][-1]
        mask = get_kmer_mask(sequences.intervals, last_end, self.k)
        kmers = self.np_get_kmers(sequences.sequences.view(self._dtype))[:mask.size]
        reverse_hashes = sequences.encoding.complement(kmers) & self._dtype(4**self.k-1)
        forward_hashes = twobit_swap(kmers) >> ((self._n_letters_in_dtype-self.k)*2)
        return forward_hashes, reverse_hashes, mask

    def np_get_kmers_with_buffer(self, sequence):
        res = (sequence[:-1, None] >> self._shifts)
        res |= (sequence[1:, None] << self._rev_shifts)
        return res
    get_kmers_with_buffer = np_get_kmers_with_buffer

    def np_get_kmers(self, sequence, is_internal=False):
        assert sequence.dtype==self._dtype, sequence.dtype
        result = (sequence[:, None] >> self._shifts)
        result[:-1] |= (sequence[1:, None] << self._rev_shifts)
        if is_internal:
            return result
        result &= self._mask
        n_kmers = self._n_letters_in_dtype*sequence.size-self.k+1
        return result.ravel()[:n_kmers]

    def get_kmers(self, sequence):
        return self.np_get_kmers(sequence, is_internal=True)

    def _has_buffer(self, sequences):
        last_end = sequences.intervals[1][-1]
        return sequences.sequences.size*4 - last_end>=self._n_letters_in_dtype-self.k+1

    def get_new_kmer_hashes(self, sequences):
        last_end = sequences.intervals[1][-1]
        mask = get_kmer_mask(sequences.intervals, last_end, self.k)
        func = self.get_kmers_with_buffer if self._has_buffer(sequences) else self.get_kmers
        kmers = func(sequences.sequences.view(self._dtype))
        out = kmers.ravel()[:mask.size][mask] & self._mask
        return out

    def get_kmer_hashes(self, sequences):
        return ACTGTwoBitEncoding.complement(self.get_new_kmer_hashes(sequences)) & self._dtype(4**self.k-1)

class KmerHash:
    CODES = np.zeros(256, dtype=np.uint64)
    CODES[[97, 65]] = 0
    CODES[[99, 67]] = 1
    CODES[[116, 84]] = 2
    CODES[[103, 71]] = 3

    def __init__(self, k=31):
        self.k=k
        self.POWER_ARRAY = np.power(4, np.arange(k, dtype=np.uint64), dtype=np.uint64)
        self.REV_POWER_ARRAY = np.power(4, np.arange(k, dtype=np.uint64)[::-1], dtype=np.uint64)

    def _to_text(self, seq):
        return "".join(letters[n] for n in seq)

    def _get_kmers(self, seq):
        return [self._to_text(seq[i:i+self.k]) for i in range(len(seq)-self.k+1)]

    def _get_codes(self, seq):
        return self.CODES[seq]

    def _join(self, kmers1, kmers2, mask):
        return np.concatenate((kmers1[mask], kmers2[mask]))

    def get_kmer_hashes(self, sequences):
        codes = self._get_codes(sequences.sequences)
        assert codes.dtype==np.uint64, codes.dtype
        kmers = np.convolve(codes, self.POWER_ARRAY, mode="valid")
        reverse_kmers = np.convolve((codes+2) % 4, self.REV_POWER_ARRAY, mode="valid")
        mask = get_kmer_mask((sequences.intervals_start, sequences.intervals_end), sequences.sequences.size, k=self.k)
        return kmers, reverse_kmers, mask # [mask], reverse_kmers[mask]



import numpy as np
#import numpy_indexed as npi
import resource
import logging

def read_fasta(file_name):
    i = 0
    with open(file_name, "rb") as f:
        for line in f:
            if line[0] != 62:
                i += 1
                yield line



def remap_array(array, from_values, to_values):
    index = np.digitize(array.ravel(), from_values, right=True)
    return to_values[index].reshape(array.shape)



def log_memory_usage_now(logplace=""):
    memory = int(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss) / 1000000
    logging.info("Memory usage (%s): %.4f GB" % (logplace, memory))



import time
import logging
import sys
from shared_memory_wrapper import SingleSharedArray, to_shared_memory, from_shared_memory
import gzip
import numpy as np

from .sequences import Sequences
from .encodings import BaseEncoding, ACTGTwoBitEncoding
NEWLINE = 10

def get_mask_from_intervals(intervals, size):
    """ intervals = (starts, ends) """
    mask_changes = np.zeros(size+1, dtype="bool")
    mask_changes[intervals[0]]=True
    mask_changes[intervals[1]]^=True
    mask = np.logical_xor.accumulate(mask_changes)
    return mask[:-1]

class FileBuffer:
    _buffer_divisor = 1
    def __init__(self, data, new_lines):
        self._data = data
        self._new_lines = new_lines
        self._is_validated = False

    @classmethod
    def from_raw_buffer(cls, chunk):
        raise NotImplemented

    @property
    def size(self):
        return self._data.size

    def _move_intervals_to_contigous_array(self, starts, ends):
        """
        Create a mask for where the sequences are and move sequences to continous array
        """
        mask = get_mask_from_intervals((starts, ends), self._data.size)
        removed_areas = np.cumsum(starts-np.insert(ends[:-1], 0, 0))
        new_intervals = (starts-removed_areas, ends-removed_areas)
        m = np.sum(mask)
        d = m % self._buffer_divisor
        seq = np.empty(m-d+self._buffer_divisor, dtype=self._data.dtype)
        seq[:m] = self._data[mask]
        return seq, new_intervals

    def validate_if_not(self):
        if not self._is_validated:
            self._validate()


class OneLineBuffer(FileBuffer):
    n_lines_per_entry = 2
    _encoding = BaseEncoding

    @classmethod
    def from_raw_buffer(cls, chunk):
        new_lines = np.flatnonzero(chunk==NEWLINE)
        n_lines = new_lines.size
        assert n_lines >= cls.n_lines_per_entry, "No complete entry in buffer"
        new_lines = new_lines[:n_lines-(n_lines%cls.n_lines_per_entry)]
        return cls(chunk[:new_lines[-1]+1], new_lines)

    def get_sequences(self):
        self.validate_if_not()
        sequence_starts = self._new_lines[::self.n_lines_per_entry]+1
        sequence_ends = self._new_lines[1::self.n_lines_per_entry]
        seq, new_intervals = self._move_intervals_to_contigous_array(sequence_starts, sequence_ends)
        seq = self._encoding.from_bytes(seq)
        return Sequences(seq, new_intervals, encoding=self._encoding)

    def _validate(self):
        n_lines = self._new_lines.size
        assert n_lines % self.n_lines_per_entry == 0, "Wrong number of lines in buffer"
        header_idxs = self._new_lines[self.n_lines_per_entry-1:-1:self.n_lines_per_entry]+1
        assert np.all(self._data[header_idxs]==self.HEADER)
        self._is_validated = True

class OneLineFastaBuffer(OneLineBuffer):
    HEADER= 62
    n_lines_per_entry = 2

class FastQBuffer(OneLineBuffer):
    n_lines_per_entry = 4
    _encoding = BaseEncoding

class OneLineFastaBuffer2Bit(OneLineFastaBuffer):
    _encoding = ACTGTwoBitEncoding
    _buffer_divisor = 32

class FastQBuffer2Bit(FastQBuffer):
    _encoding = ACTGTwoBitEncoding
    _buffer_divisor = 32

class BufferedNumpyParser:

    def __init__(self, file_obj, buffer_type, chunk_size=1000000):
        self._file_obj = file_obj
        self._chunk_size = chunk_size
        self._is_finished = False
        self._buffer_type = buffer_type

    @classmethod
    def from_filename(cls, filename, *args, **kwargs):
        if any(filename.endswith(suffix) for suffix in ("fasta", "fa", "fasta.gz", "fa.gz")):
            buffer_type = OneLineFastaBuffer2Bit
        elif any(filename.lower().endswith(suffix) for suffix in ("fastq", "fq", "fastq.gz", "fq.gz")):
            buffer_type = FastQBuffer2Bit
        else:
            raise NotImplemented
        if filename.endswith(".gz"):
            file_obj = gzip.open(filename, "rb")
        else:
            file_obj = open(filename, "rb")
        return cls(file_obj, buffer_type, *args, **kwargs)

    def get_chunk(self):
        a, bytes_read = self.read_raw_chunk()
        self._is_finished = bytes_read < self._chunk_size
        if bytes_read == 0:
            return None

        # Ensure that the last entry ends with newline. Makes logic easier later
        if self._is_finished  and a[bytes_read-1] != NEWLINE:
            a[bytes_read] = NEWLINE
            bytes_read += 1
        return a[:bytes_read]

    def read_raw_chunk(self):
        array = np.empty(self._chunk_size, dtype="uint8")
        bytes_read = self._file_obj.readinto(array)
        return array, bytes_read

    def get_chunks(self):
        chunk = self.get_chunk()
        while not self._is_finished:
            t = time.perf_counter()
            buff = self._buffer_type.from_raw_buffer(chunk)
            self._file_obj.seek(buff.size-self._chunk_size, 1)
            logging.info("Spent %.4f sec reading chunk from file" % (time.perf_counter()-t))
            yield buff
            chunk = self.get_chunk()
        if chunk is not None and chunk.size:
            yield self._buffer_type.from_raw_buffer(chunk)

class TextParser:
    _encoding = BaseEncoding

    def __init__(self, filename, chunk_size=1000000):
        self._file_obj = self._get_file_obj(filename)
        self._chunk_size = chunk_size
        self._is_finished = False
        self._last_header_idx = -1

    def _get_file_obj(self, filename):
        if filename.endswith(".gz"):
            return gzip.open(filename, "rb")
        return open(filename, "rb")

    def _cut_array(self, array, last_newline):
        self._last_header_idx = last_newline+1
        return array[:last_newline]

    def read_raw_chunk(self):
        array = np.empty(self._chunk_size, dtype="uint8")
        bytes_read = self._file_obj.readinto(array)
        return array, bytes_read

    def parse(self, as_shared_memory_object=False):
        while not self._is_finished:
            t = time.perf_counter()
            chunk = self.parse_chunk()
            logging.info("Time to parse chunk: %.4f" % (time.perf_counter()-t))
            if chunk is not None:
                if as_shared_memory_object:
                    shared_memory_name = str(np.random.randint(0, 10e15))
                    to_shared_memory(chunk, shared_memory_name)
                    yield shared_memory_name
                else:
                    yield chunk

    def _handle_chunk(self, chunk):
        return self.get_sequences(chunk)

    def get_raw_chunks(self, as_shared_memory=False):
        raw_chunk = self.read_raw_chunk()
        while not self._is_finished and raw_chunk:
            if as_shared_memory_object:
                shared_memory_name = str(np.random.randint(0, 10e15))
                to_shared_memory(chunk, shared_memory_name)
                yield shared_memory_name
            else:
                yield chunk

    def cut_chunk(self, chunk):
        new_lines = np.flatnonzero(chunk == NEW_LINE)
        return self._cut_array(chunk, new_lines[-1])


    def parse_chunk(self):
        a, bytes_read = self.read_raw_chunk()
        self._is_finished = bytes_read < self._chunk_size
        if bytes_read == 0:
            return None
        if self._is_finished  and a[bytes_read-1] != NEWLINE:
            a[bytes_read]  = NEWLINE
        sequences = self._handle_chunk(a[:bytes_read])
        if not self._is_finished:
            self._file_obj.seek(self._last_header_idx-self._chunk_size, 1)
        return sequences

class OneLineParser(TextParser):
    """
    Base class for formats where the sequnences
    are constrained to sinlge line
    """
    _buffer_divisor=1

    def cut_chunk(self, chunk):
        new_lines = np.flatnonzero(chunk == NEW_LINE)
        return self._cut_array(chunk, new_lines[-1])

    def get_sequences(self, raw_chunk):
        new_lines = np.flatnonzero(raw_chunk==NEWLINE)
        chunk, new_lines = self._validate_chunk(raw_chunk, new_lines)
        sequence_starts = new_lines[::self.n_lines_per_entry]+1
        sequence_ends = new_lines[1::self.n_lines_per_entry]
        return self._mask_and_move_sequences(chunk, sequence_starts, sequence_ends)

    def _mask_and_move_sequences(self, array, sequence_starts, sequence_ends):
        """
        Create a mask for where the sequences are and move sequences to continous array
        """
        mask = get_mask_from_intervals((sequence_starts, sequence_ends), array.size)
        removed_areas = np.cumsum(sequence_starts-np.insert(sequence_ends[:-1], 0, 0))
        new_intervals = (sequence_starts-removed_areas, sequence_ends-removed_areas)
        m = np.sum(mask)
        d = m%self._buffer_divisor
        seq = np.empty(m-d+self._buffer_divisor, dtype=array.dtype)
        seq[:m] = array[mask]
        seq = self._encoding.from_bytes(seq)
        return Sequences(seq, new_intervals, encoding=self._encoding)

    def _validate_chunk(self, chunk, new_lines):
        assert chunk[0] == self.HEADER, "Chunk does not start with header"
        n_lines = new_lines.size
        assert n_lines >=self.n_lines_per_entry, "No complete entry in buffer"
        new_lines = new_lines[:n_lines-(n_lines%self.n_lines_per_entry)]
        header_idxs = new_lines[self.n_lines_per_entry-1:-1:self.n_lines_per_entry]+1
        assert np.all(chunk[header_idxs]==self.HEADER)
        return self._cut_array(chunk, new_lines[-1]), new_lines

class FastqParser(OneLineParser):
    HEADER = 64
    n_lines_per_entry = 4


class FastqParser2Bit(FastqParser):
    _buffer_divisor = 32
    _encoding = ACTGTwoBitEncoding


class OneLineFastaParser(OneLineParser):
    HEADER= 62
    n_lines_per_entry = 2


class LooseOneLineFastaParser(OneLineParser):
    HEADER = 62
    n_lines_per_entry = 2

    def _validate_chunk(self, chunk, new_lines):
        assert chunk[0]==self.HEADER

    def get_sequences(self, array):
        new_lines = np.flatnonzero(array==NEWLINE)
        # Find which lines are sequences, and cut the array so that it ends with a complete sequence
        is_sequence_line = array[new_lines[:-1]+1] != self.HEADER
        # assert np.any(is_sequence_line), (array, self._is_finished)
        idx_last_sequence_line = np.flatnonzero(is_sequence_line)[-1]
        new_lines = new_lines[:idx_last_sequence_line+2]
        array = self._cut_array(array, new_lines[-1])
        is_sequence_line = is_sequence_line[:idx_last_sequence_line+1]
        sequence_starts = new_lines[:-1][is_sequence_line]+1
        sequence_ends = new_lines[1:][is_sequence_line]
        return self._mask_and_move_sequences(array, sequence_starts, sequence_ends)

    def _mask_and_move_sequences__(self, array, sequence_starts, sequence_ends):
        """
        Create a mask for where the sequences are and move sequences to continous array
        """
        mask = get_mask_from_intervals((sequence_starts, sequence_ends), array.size)
        removed_areas = np.cumsum(sequence_starts-np.insert(sequence_ends[:-1], 0, 0))
        new_intervals = (sequence_starts-removed_areas, sequence_ends-removed_areas)
        return Sequences(array[mask], new_intervals[0], new_intervals[1])


class OneLineFastaParser2bit(OneLineFastaParser):
    _buffer_divisor = 32
    _encoding = ACTGTwoBitEncoding
    def _mask_and_move_sequences__(self, array, sequence_starts, sequence_ends):
        """
        Create a mask for where the sequences are and move sequences to continous array
        """
        mask = get_mask_from_intervals((sequence_starts, sequence_ends), array.size)
        removed_areas = np.cumsum(sequence_starts-np.insert(sequence_ends[:-1], 0, 0))
        new_intervals = (sequence_starts-removed_areas, sequence_ends-removed_areas)
        m = np.sum(mask)
        d = m%32
        seq = np.empty(m-d+32, dtype=array.dtype)
        seq[:m] = array[mask]
        return Sequences(seq, new_intervals[0], new_intervals[1])


class FastaParser(TextParser):
    def _get_sequence_offsets(self, new_lines, is_header):
        header_starts = new_lines[:-1][is_header[:-1]]
        header_ends = new_lines[1:][is_header[:-1]]
        header_lens = header_ends-header_starts
        masked_count = np.ones(new_lines.size, dtype="int")
        masked_count[1:][is_header[:-1]] = header_lens
        masked_count[0] = new_lines[0]+1
        cum_masked_count = np.cumsum(masked_count)
        indexes_after_masking = new_lines+1-cum_masked_count
        starts = indexes_after_masking[1:][is_header[:-1]]
        if not self._is_finished:
            starts = starts[:-1]
        return starts

    def get_sequences(self, array):
        assert array[0] == HEADER
        all_new_lines = np.flatnonzero(array==NEWLINE)
        new_lines = all_new_lines if all_new_lines[-1] != array.size-1 else all_new_lines[:-1]
        is_header = array[new_lines+1] == HEADER
        mask_changes = np.zeros(array.size, dtype="bool")
        mask_changes[new_lines+1] = np.where(is_header, 1, 0)
        mask_changes[new_lines[1:]] = np.where(is_header[:-1], 1, 0)
        mask_changes[0] = 1
        mask_changes[new_lines[0]] = 1
        mask = np.logical_xor.accumulate(mask_changes)
        mask[all_new_lines] = 1
        beginnings = self._get_sequence_offsets(new_lines, is_header)
        intervals = (np.insert(beginnings, 0, 0), np.append(beginnings, np.count_nonzero(~mask)))
        if self._is_finished:
            return Sequences(array[~mask], intervals)
        assert np.any(is_header), ("Fasta entry to big for chunk size", to_text(array), chunk_size)
        self._last_header_idx = new_lines[is_header][-1]+1
        intervals = (np.insert(beginnings, 0, 0), np.append(beginnings, np.count_nonzero(~mask[:self._last_header_idx])))
        return Sequences(array[:self._last_header_idx][~mask[:self._last_header_idx]], intervals)



import gzip
import logging
import os
import sys
import time
from pathlib import PurePath
import tqdm

logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s')

import numpy as np
import warnings
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 

import pyximport
pyximport.install()

from graph_kmer_index.collision_free_kmer_index import MinimalKmerIndex, CounterKmerIndex
import argparse
from graph_kmer_index import KmerIndex
from .mapping import map_fasta
from graph_kmer_index.index_bundle import IndexBundle
from shared_memory_wrapper.shared_memory import remove_shared_memory_in_session, remove_shared_memory
from shared_memory_wrapper.shared_memory import get_shared_pool
from shared_memory_wrapper import from_file, object_to_shared_memory, object_from_shared_memory
import bionumpy as bnp
from kmer_mapper.mapper import map_kmers_to_graph_index
from bionumpy.encodings import ACTGTwoBitEncoding, ACTGEncoding
from shared_memory_wrapper.shared_array_map_reduce import additative_shared_array_map_reduce
import npstructures


# modified version of bionumpy's fast_hash
# will use bionumpy's fast hash in the future
@bnp.util.convolution
def fast_hash(sequence, k, encoding=None):
    sequence = bnp.sequences.as_encoded_sequence_array(sequence, ACTGEncoding)
    if encoding:
        sequence = encoding.encode(sequence)

    bit_array = npstructures.bitarray.BitArray.pack(sequence, bit_stride=2)
    hashes = bit_array.sliding_window(k)
    return hashes




def main():
    run_argument_parser(sys.argv[1:])


def _get_kmer_index_from_args(args):
    # allowing multiple ways to specify kmer index, try to get the index
    if args.kmer_index is None:
        if args.index_bundle is None:
            logging.error("Either a kmer index (-i) or an index bundle (-b) needs to be specified")
            sys.exit(1)
        else:
            kmer_index = IndexBundle.from_file(args.index_bundle).indexes["KmerIndex"]
            kmer_index.convert_to_int32()
            kmer_index.remove_ref_offsets()  # not needed, will save us some memory
    else:

        cls = KmerIndex
        if "minimal" in args.kmer_index:
            cls = MinimalKmerIndex
        try:
            kmer_index = cls.from_file(args.kmer_index)
            kmer_index.convert_to_int32()
            kmer_index.remove_ref_offsets()  # not needed, will save us some memory
        except:
            kmer_index = from_file(args.kmer_index)
            assert isinstance(kmer_index, CounterKmerIndex)
            logging.info("Kmer index is counter index")

    return kmer_index


def map_fasta_command(args):
    if args.kmer_size > 31:
        logging.error("k must be 31 or lower")
        sys.exit(1)

    logging.info("Max read length is specified to %d" % args.max_read_length)
    logging.info("Max hits per kmer: %d" % args.max_hits_per_kmer)

    if not args.fasta_file.endswith(".fa"):
        logging.error("Only fasta files (not fq or gzipped files) are supported to the argument -f. If you have another"
                          "format, you can pipe fasta to kmer_mapper and use a dash as file name (-f -)")
        sys.exit(1)

    get_shared_pool(args.n_threads)
    kmer_index = None

    kmer_index = _get_kmer_index_from_args(args)

    args.max_node_id = kmer_index.max_node_id()
    logging.info("Max node id is %d" % args.max_node_id)
    map_fasta(args, kmer_index)


def _mapper(kmer_size, kmer_index, chunk_sequence_name):
    t = time.perf_counter()
    chunk_sequence = object_from_shared_memory(chunk_sequence_name).get_sequences()
    hashes = fast_hash(chunk_sequence, kmer_size, encoding=None).ravel()
    hashes = ACTGTwoBitEncoding.complement(hashes) & np.uint64(4 ** kmer_size - 1)

    t0 = time.perf_counter()
    if isinstance(kmer_index, CounterKmerIndex):
        mapped = kmer_index.counter.count(hashes, return_counts=True)._values
        logging.debug("Mapped with counter. Got values of length %d" % len(mapped))
    else:
        mapped = map_kmers_to_graph_index(kmer_index, kmer_index.max_node_id(), hashes)

    logging.debug("Mapping %d hashes took %.3f sec" % (len(hashes), time.perf_counter()-t0))

    remove_shared_memory(chunk_sequence_name)
    logging.debug("Chunk of %d reads took %.2f sec" % (len(chunk_sequence), time.perf_counter()-t))
    return mapped


def open_file(filename):
    path = PurePath(filename)
    suffix = path.suffixes[-1]
    try:
        buffer_type = bnp.files._get_buffer_type(suffix)
    except RuntimeError:
        logging.error("Unsupported file suffix %s" % suffix)
        raise
    
    if path.suffixes[-1] == ".fa":
        # override buffer type for some performance gain
        buffer_type = bnp.TwoLineFastaBuffer
        logging.info("Using buffer type TwoLineFastaBuffer")

    open_func = gzip.open if suffix == ".gz" else open
    return bnp.parser.NumpyFileReader(open_func(filename, "rb"), buffer_type)


def map_bnp(args):
    if args.debug:
        logging.info("Will print debug log")
        logging.getLogger().setLevel(logging.DEBUG)

    k = args.kmer_size
    kmer_index = _get_kmer_index_from_args(args)

    start_time = time.perf_counter()

    file = open_file(args.reads)

    n_bytes = os.stat(args.reads).st_size
    approx_number_of_chunks = int(n_bytes / args.chunk_size)

    chunks = tqdm.tqdm((object_to_shared_memory(raw_chunk) for
        raw_chunk in file.read_chunks(chunk_size=args.chunk_size)), total=approx_number_of_chunks)

    if isinstance(kmer_index, KmerIndex):
        initial_data = np.zeros(kmer_index.max_node_id()+1)
    else:
        initial_data = np.zeros_like(kmer_index.counter._values)

    node_counts = additative_shared_array_map_reduce(_mapper,
                                                     chunks,
                                                     initial_data,
                                                     (args.kmer_size, kmer_index),
                                                     n_threads=args.n_threads
                                                     )

    if isinstance(kmer_index, CounterKmerIndex):
        # node counts is not node counts, but kmer counts, get node counts
        t = time.perf_counter()
        kmer_index.counter._values = node_counts
        node_counts = kmer_index.get_node_counts()
        logging.info("Time spent getting node counts in the end: %.3f" % (time.perf_counter()-t))


    np.save(args.output_file, node_counts)
    logging.info("Saved node counts to %s.npy" % args.output_file)
    logging.info("Spent %.3f sec in total mapping kmers using %d threads" % (time.perf_counter()-start_time, args.n_threads))
    remove_shared_memory_in_session()



def run_argument_parser(args):

    parser = argparse.ArgumentParser(
        description='Kmer Mapper',
        prog='kmer_mapper',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50, width=100))

    subparsers = parser.add_subparsers()
    subparser = subparsers.add_parser("legacy_map", help="Legacy mapper kept for backwards compatibility")
    subparser.add_argument("-i", "--kmer-index", required=False)
    subparser.add_argument("-b", "--index-bundle", required=False)
    subparser.add_argument("-f", "--fasta-file", required=True)
    subparser.add_argument("-k", "--kmer-size", required=False, default=31, type=int)
    subparser.add_argument("-t", "--n-threads", required=False, default=16, type=int)
    subparser.add_argument("-c", "--chunk-size", required=False, type=int, default=500000, help="N reads to process in each chunk")
    subparser.add_argument("-l", "--max-read-length", required=False, type=int, default=150,
                           help="Maximum length of reads. Reads should not be longer than this.")
    subparser.add_argument("-o", "--output-file", required=True)
    subparser.add_argument("-n", "--use-numpy", required=False, type=bool, default=False, help="Use numpy-based counting instead of Cython")
    #subparser.add_argument("-N", "--use-cython-file-reading", required=False, type=bool, default=False, help="Use cython-based file reading instead of Cython")
    #subparser.add_argument("-m", "--no-shared-memory", required=False, type=bool, default=False, help="Set to True to not use shared memory for index. Increases memory usage by a factor of --n-threads.")
    #subparser.add_argument("-r", "--include-reverse-complement", required=False, default=False, type=bool)
    #subparser.add_argument("-T", "--use-numpy-parsing", required=False, default=False, type=bool)
    subparser.add_argument("-I", "--max-hits-per-kmer", required=False, default=1000, type=int,
                           help="Ignore kmers that have more than this amount of hits in index")
    subparser.set_defaults(func=map_fasta_command)



    subparser = subparsers.add_parser("map", help="Map reads to a kmer index")
    subparser.add_argument("-i", "--kmer-index", required=False)
    subparser.add_argument("-b", "--index-bundle", required=False)
    subparser.add_argument("-f", "--reads", required=True, help="Reads in .fa, .fq, .fa.gz, or fq.gz format")
    subparser.add_argument("-k", "--kmer-size", required=False, default=31, type=int)
    subparser.add_argument("-t", "--n-threads", required=False, default=16, type=int)
    subparser.add_argument("-c", "--chunk-size", required=False, type=int, default=100000000,
                           help="N bytes to process in each chunk")
    subparser.add_argument("-o", "--output-file", required=True)
    subparser.add_argument("-d", "--debug", required=False, help="Set to True to print debug log")
    subparser.add_argument("-I", "--max-hits-per-kmer", required=False, default=1000, type=int,
                           help="Ignore kmers that have more than this amount of hits in index")
    subparser.set_defaults(func=map_bnp)


    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args)
    args.func(args)

    remove_shared_memory_in_session()
